#include "winver.h"
#ifdef LOCALIZE_VERSION_INFO
  #pragma code_page (65001)
#endif

VS_VERSION_INFO VERSIONINFO
  FILEVERSION @rime_version_comma_separated@,0
  PRODUCTVERSION @rime_version_comma_separated@,0
  FILEFLAGSMASK 0x17L
#ifdef _DEBUG
  FILEFLAGS 0x1L
#else
  FILEFLAGS 0x0L
#endif
  FILEOS 0x40004L
  FILETYPE 0x1L
  FILESUBTYPE 0x0L
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
#ifdef LOCALIZE_VERSION_INFO
    BLOCK "080404b0"
    BEGIN
      VALUE "CompanyName", "式恕堂"
      VALUE "FileDescription", "中州韵输入法引擎"
      VALUE "InternalName", "librime"
      VALUE "LegalCopyright", "式恕堂 版权所无"
      VALUE "OriginalFilename", "@original_filename@"
      VALUE "ProductName", "中州韵输入法引擎"
      VALUE "ProductVersion", "@rime_version@@tag_suffix@"
    END
    BLOCK "040904b0"
    BEGIN
      VALUE "CompanyName", "式恕堂"
      VALUE "FileDescription", "中州韻輸入法引擎"
      VALUE "InternalName", "librime"
      VALUE "LegalCopyright", "式恕堂 版權所無"
      VALUE "OriginalFilename", "@original_filename@"
      VALUE "ProductName", "中州韻輸入法引擎"
      VALUE "ProductVersion", "@rime_version@@tag_suffix@"
    END
#endif
    BLOCK "040904E4"
    BEGIN
      VALUE "CompanyName", "Rime Developers"
      VALUE "FileDescription", "Rime Input Method Engine"
      VALUE "InternalName", "librime"
      VALUE "LegalCopyright", "Copyleft Rime Developers"
      VALUE "OriginalFilename", "@original_filename@"
      VALUE "ProductName", "Rime Input Method Engine"
      VALUE "ProductVersion", "@rime_version@@tag_suffix@"
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
#ifdef LOCALIZE_VERSION_INFO
    VALUE "Translation", 0x804, 1200
    VALUE "Translation", 0x409, 1200
#endif
    VALUE "Translation", 0x409, 1252
  END
END
